from estimators.ccb import base
from math import inf
from typing import List, Callable
from estimators.bandits.cressieread import EstimatorImpl, IntervalImpl
from copy import deepcopy


class Estimator(base.Estimator):
    wmin: float
    wmax: float
    _impl: List[EstimatorImpl]

    def __init__(self, wmin: float = 0, wmax: float = inf):
        self.wmin = wmin
        self.wmax = wmax
        self._impl = []

    def add_example(self, p_logs: List[float], rs: List[float], p_preds: List[float], count: float = 1.0) -> None:
        if count > 0:
            ws = [p_pred / p_log for p_pred, p_log in zip(p_preds, p_logs)]
            w = 1.0
            for i in range(len(ws)):
                w *= ws[i]
                if len(self._impl) <= i:
                    self._impl.append(EstimatorImpl(self.wmin ** (i + 1), self.wmax ** (i + 1)))
                self._impl[i].add(w, rs[i], count)

    def get(self) -> List[float]:
        result = []
        n0 = float(self._impl[0].n) if any(self._impl) else 0
        if n0 > 0:
            for impl in self._impl:
                result.append(impl.get() * float(impl.n) / n0)
        return result

    def __add__(self, other: 'Estimator') -> 'Estimator':
        (large, small) = (self, other) if len(self._impl) >= len(other._impl) else (other, self)
        result = Estimator(wmin = min(self.wmin, other.wmin), wmax = max(self.wmax, other.wmax))
        for i in range(len(large._impl)):
            result._impl.append(large._impl[i] + small._impl[i] if i < len(small._impl) else deepcopy(large._impl[i]))
        return result


class Interval(base.Interval):
    wmin: float
    wmax: float
    rmin: float
    rmax: float
    _impl: List[IntervalImpl]

    def __init__(self, wmin: float = 0, wmax: float = inf, rmin: float = 0, rmax: float = 1, empirical_r_bounds: bool = False):
        self.wmin = wmin
        self.wmax = wmax
        self.rmin = rmin
        self.rmax = rmax
        self._impl = []
        self.empirical_r_bounds = empirical_r_bounds

    def add_example(self, p_logs: List[float], rs: List[float], p_preds: List[float], count: float = 1.0) -> None:
        if count > 0:
            ws = [p_pred / p_log for p_pred, p_log in zip(p_preds, p_logs)]
            w = 1.0
            for i in range(len(ws)):
                w *= ws[i]
                if len(self._impl) <= i:
                    self._impl.append(IntervalImpl(self.wmin ** (i + 1), self.wmax ** (i + 1), self.rmin, self.rmax, self.empirical_r_bounds))
                self._impl[i].add(w, rs[i], count)

    def get(self, alpha: float = 0.05, atol: float = 1e-9) -> List[List[float]]:
        result = []
        n0 = float(self._impl[0].n) if any(self._impl) else 0
        if n0 > 0:
            for impl in self._impl:
                result.append([v * float(impl.n) / n0 for v in impl.get()])
        return result

    def __add__(self, other: 'Interval') -> 'Interval':
        assert not (self.empirical_r_bounds ^ other.empirical_r_bounds), 'Summation of estimators with various r bounds policy is prohibited'
        
        if not self.empirical_r_bounds:
            assert self.rmin == other.rmin, 'Summation of estimators with various r bounds is prohibited'
            assert self.rmax == other.rmax, 'Summation of estimators with various r bounds is prohibited'

        (large, small) = (self, other) if len(self._impl) >= len(other._impl) else (other, self)
        result = Interval(
            wmin = min(self.wmin, other.wmin),
            wmax=max(self.wmax, other.wmax),
            rmin=min(self.rmin, other.rmin),
            rmax=max(self.rmax, other.rmax),
            empirical_r_bounds=self.empirical_r_bounds)
        for i in range(len(large._impl)):
            result._impl.append(large._impl[i] + small._impl[i] if i < len(small._impl) else deepcopy(large._impl[i]))
        return result



from scipy.stats import beta
from estimators.bandits import base
from typing import List, Optional
from estimators.math import clopper_pearson


class Interval(base.Interval):
    examples_count: float
    weighted_reward: float
    max_weight: float

    def __init__(self, rmin: float = 0, rmax: float = 1):
        self.examples_count = 0
        self.weighted_reward = 0
        self.max_weight = 0
        self.rmin = rmin
        self.rmax = rmax

    def _scale(self, r):
        assert r >= self.rmin and r <= self.rmax, f'Error: {r} is out of [{self.rmin}, {self.rmax}]'
        return (r - self.rmin) / (self.rmax - self.rmin)

    def _scale_back(self, r):
        return self.rmin + r * (self.rmax - self.rmin)

    def add_example(self, p_log: float, r: float, p_pred: float, count: float = 1.0) -> None:
        r = self._scale(r)
        self.examples_count += count
        w = p_pred / p_log
        self.weighted_reward += r * w * count
        self.max_weight = max(self.max_weight, w)

    def get(self, alpha: float = 0.05) -> List[Optional[float]]:
        if self.max_weight > 0.0:
            successes = self.weighted_reward / self.max_weight
            n = self.examples_count / self.max_weight
            cp = clopper_pearson(successes, n, alpha)
            return [self._scale_back(cp[0]), self._scale_back(cp[1])]
        return [None, None]

    def __add__(self, other: 'Interval') -> 'Interval':
        result = Interval()
        result.examples_count = self.examples_count + other.examples_count
        result.weighted_reward = self.weighted_reward + other.weighted_reward
        result.max_weight = max(self.max_weight, other.max_weight)
        return result




import argparse
import math


def set_custom_args(parser):
    parser.add_argument('--max_value', help="[CATS estimator] max value for continuous action range", required=False)
    parser.add_argument('--min_value', help="[CATS estimator] min value for continuous action range", required=False)
    parser.add_argument('--num_actions', help="[CATS estimator] number of actions used to discretize continuous range",
                        required=False)
    parser.add_argument('--bandwidth', help="[CATS estimator] bandwidth (radius) of randomization around discrete "
                                            "actions in terms of continuous range ", required=False)
    return parser


def get_cats_transformer(args):
    if args.num_actions and args.max_value and args.min_value and args.bandwidth:
        return CatsTransformer(args.num_actions, args.bandwidth, args.max_value, args.min_value)
    else:
        return


class CatsTransformer:
    def __init__(self, num_actions, bandwidth, max_value, min_value):
        self.num_actions = int(num_actions)
        self.max_value = float(max_value)
        self.min_value = float(min_value)
        self.bandwidth = float(bandwidth)

        self.continuous_range = self.max_value - self.min_value
        self.unit_range = self.continuous_range / float(self.num_actions)
    
    def get_baseline1_prediction(self):
        return self.min_value + (self.unit_range / 2.0)
    
    def transform(self, data, pred_a):        
        logged_a = data['a']

        ctr = min((self.num_actions - 1), math.floor((pred_a - self.min_value) / self.unit_range))
        centre = self.min_value + ctr * self.unit_range + (self.unit_range / 2.0)

        if math.isclose(centre, logged_a, abs_tol=self.bandwidth):
            b = min(self.max_value, centre + self.bandwidth) - max(self.min_value, centre - self.bandwidth)
            data['pred_p'] = 1.0 / b
        else:
            data['pred_p'] = 0.0
        
        return data



from estimators.bandits import base
from typing import Optional


class Estimator(base.Estimator):
    weighted_examples_count: float
    weighted_reward: float

    def __init__(self):
        self.weighted_examples_count = 0
        self.weighted_reward = 0

    def add_example(self, p_log: float, r: float, p_pred: float, count: float = 1.0) -> None:
        w = p_pred / p_log
        self.weighted_examples_count += w * count
        self.weighted_reward += r * w * count

    def get(self) -> Optional[float]:
        return self.weighted_reward / self.weighted_examples_count if self.weighted_examples_count != 0 else None

    def __add__(self, other: 'Estimator') -> 'Estimator':
        result = Estimator()
        result.weighted_examples_count = self.weighted_examples_count + other.weighted_examples_count
        result.weighted_reward = self.weighted_reward + other.weighted_reward
        return result



# CR(-2) is particularly computationally convenient

from math import inf
from estimators.bandits import base
from typing import List, Optional
from estimators.math import IncrementalFsum


class EstimatorImpl:
    wmin: float
    wmax: float
    n: IncrementalFsum
    sumw: IncrementalFsum
    sumwsq: IncrementalFsum
    sumwr: IncrementalFsum
    sumwsqr: IncrementalFsum
    sumr: IncrementalFsum

    # NB: This works better you use the true wmin and wmax
    #     which is _not_ the empirical minimum and maximum
    #     but rather the actual smallest and largest possible values

    def __init__(self, wmin: float = 0, wmax: float = inf):
        assert wmin < 1
        assert wmax > 1

        self.wmin = wmin
        self.wmax = wmax

        self.n = IncrementalFsum()
        self.sumw = IncrementalFsum()
        self.sumwsq = IncrementalFsum()
        self.sumwr = IncrementalFsum()
        self.sumwsqr = IncrementalFsum()
        self.sumr = IncrementalFsum()

    def add(self, w: float, r: float, count: float = 1.0) -> None:
        if count > 0:
            assert w >= 0, 'Error: negative importance weight'

            self.n += count
            self.sumw += count * w
            self.sumwsq += count * w ** 2
            self.sumwr += count * w * r
            self.sumwsqr += count * w ** 2 * r
            self.sumr += count * r

            self.wmax = max(self.wmax, w)
            self.wmin = min(self.wmin, w)

    def get(self) -> Optional[float]:
        n = float(self.n)
        if n == 0:
            return None

        sumw = float(self.sumw)
        sumwsq = float(self.sumwsq)
        sumwr = float(self.sumwr)
        sumwsqr = float(self.sumwsqr)
        sumr = float(self.sumr)

        wfake = self.wmax if sumw < n else self.wmin

        if wfake == inf:
            gamma = -(1 + n) / n
            beta = 0
        else:
            a = (wfake + sumw) / (1 + n)
            b = (wfake ** 2 + sumwsq) / (1 + n)
            assert a * a < b
            gamma = (b - a) / (a * a - b)
            beta = (1 - a) / (a * a - b)

        vhat = (-gamma * sumwr - beta * sumwsqr) / (1 + n)
        missing = max(0.0, 1 - (-gamma * sumw - beta * sumwsq) / (1 + n))
        rhatmissing = sumr / n
        vhat += missing * rhatmissing

        return vhat

    def __add__(self, other: 'EstimatorImpl') -> 'EstimatorImpl':
        result = EstimatorImpl(
            wmin=min(self.wmin, other.wmin),
            wmax=max(self.wmax, other.wmax))

        result.n = self.n + other.n
        result.sumw = self.sumw + other.sumw
        result.sumwsq = self.sumwsq + other.sumwsq
        result.sumwr = self.sumwr + other.sumwr
        result.sumwsqr = self.sumwsqr + other.sumwsqr
        result.sumr = self.sumr + other.sumr

        return result


class IntervalImpl:
    wmin: float
    wmax: float
    rmin: float
    rmax: float
    n: IncrementalFsum
    sumw: IncrementalFsum
    sumwsq: IncrementalFsum
    sumwr: IncrementalFsum
    sumwsqr: IncrementalFsum
    sumwsqrsq: IncrementalFsum
    empirical_r_bounds: bool

    # NB: This works better you use the true wmin and wmax
    #     which is _not_ the empirical minimum and maximum
    #     but rather the actual smallest and largest possible values

    def __init__(self, wmin: float, wmax: float, rmin: float, rmax: float, empirical_r_bounds: bool):
        assert wmin < 1
        assert wmax > 1

        self.wmin = wmin
        self.wmax = wmax
        self.rmin = rmin
        self.rmax = rmax
        self.empirical_r_bounds = empirical_r_bounds

        self.n = IncrementalFsum()
        self.sumw = IncrementalFsum()
        self.sumwsq = IncrementalFsum()
        self.sumwr = IncrementalFsum()
        self.sumwsqr = IncrementalFsum()
        self.sumwsqrsq = IncrementalFsum()

    def add(self, w: float, r: float, count: float = 1.0) -> None:
        if count > 0:
            assert w >= 0, 'Error: negative importance weight'

            self.n += count
            self.sumw += count * w
            self.sumwsq += count * w ** 2
            self.sumwr += count * w * r
            self.sumwsqr += count * w ** 2 * r
            self.sumwsqrsq += count * w ** 2 * r ** 2

            self.wmax = max(self.wmax, w)
            self.wmin = min(self.wmin, w)

            if self.empirical_r_bounds:
                self.rmax = max(self.rmax, r)
                self.rmin = min(self.rmin, r)
            else:
                if r > self.rmax or r < self.rmin:
                    raise ValueError(f'Error: Value of r={r} is outside rmin={self.rmin}, rmax={self.rmax} bounds')

    def get(self, alpha: float = 0.05, atol: float = 1e-9) -> List[Optional[float]]:
        from math import isclose, sqrt
        from scipy.stats import f

        n = float(self.n)
        if n == 0:
            return [None, None]

        sumw = float(self.sumw)
        sumwsq = float(self.sumwsq)
        sumwr = float(self.sumwr)
        sumwsqr = float(self.sumwsqr)
        sumwsqrsq = float(self.sumwsqrsq)

        uncwfake = self.wmax if sumw < n else self.wmin
        if uncwfake == inf:
            uncgstar = 1 + 1 / n
        else:
            unca = (uncwfake + sumw) / (1 + n)
            uncb = (uncwfake ** 2 + sumwsq) / (1 + n)
            uncgstar = (1 + n) * (unca - 1) ** 2 / (uncb - unca * unca)

        delta = f.isf(q=alpha, dfn=1, dfd=n)
        phi = (-uncgstar - delta) / (2 * (1 + n))

        bounds = []
        for r, sign in ((self.rmin, 1), (self.rmax, -1)):
            candidates = []
            for wfake in (self.wmin, self.wmax):
                if wfake == inf:
                    x = sign * (r + (sumwr - sumw * r) / n)
                    y = ((r * sumw - sumwr) ** 2 / (n * (1 + n))
                         - (r ** 2 * sumwsq - 2 * r * sumwsqr + sumwsqrsq) / (1 + n)
                         )
                    z = phi + 1 / (2 * n)
                    if isclose(y * z, 0, abs_tol=1e-9):
                        y = 0

                    if isclose(y * z, 0, abs_tol=atol * atol):
                        gstar = x - sqrt(2) * atol
                        candidates.append(gstar)
                    elif z <= 0 and y * z >= 0:
                        gstar = x - sqrt(2 * y * z)
                        candidates.append(gstar)
                else:
                    barw = (wfake + sumw) / (1 + n)
                    barwsq = (wfake * wfake + sumwsq) / (1 + n)
                    barwr = sign * (wfake * r + sumwr) / (1 + n)
                    barwsqr = sign * (wfake * wfake * r + sumwsqr) / (1 + n)
                    barwsqrsq = (wfake * wfake * r * r + sumwsqrsq) / (1 + n)

                    if barwsq > barw ** 2:
                        x = barwr + ((1 - barw) * (barwsqr - barw * barwr) / (barwsq - barw ** 2))
                        y = (barwsqr - barw * barwr) ** 2 / (barwsq - barw ** 2) - (barwsqrsq - barwr ** 2)
                        z = phi + (1 / 2) * (1 - barw) ** 2 / (barwsq - barw ** 2)

                        if isclose(y * z, 0, abs_tol=atol * atol):
                            gstar = x - sqrt(2) * atol
                            candidates.append(gstar)
                        elif z <= 0 and y * z >= 0:
                            gstar = x - sqrt(2 * y * z)
                            candidates.append(gstar)

            best = min(candidates)
            vbound = min(self.rmax, max(self.rmin, sign * best))
            bounds.append(vbound)

        return bounds

    def __add__(self, other: 'IntervalImpl') -> 'IntervalImpl':
        assert not (self.empirical_r_bounds ^ other.empirical_r_bounds), 'Summation of estimators with various r bounds policy is prohibited'
        
        if not self.empirical_r_bounds:
            assert self.rmin == other.rmin, 'Summation of estimators with various r bounds is prohibited'
            assert self.rmax == other.rmax, 'Summation of estimators with various r bounds is prohibited'

        result = IntervalImpl(
            wmin=min(self.wmin, other.wmin),
            wmax=max(self.wmax, other.wmax),
            rmin=min(self.rmin, other.rmin),
            rmax=max(self.rmax, other.rmax),
            empirical_r_bounds=self.empirical_r_bounds 
        )

        result.n = self.n + other.n
        result.sumw = self.sumw + other.sumw
        result.sumwsq = self.sumwsq + other.sumwsq
        result.sumwr = self.sumwr + other.sumwr
        result.sumwsqr = self.sumwsqr + other.sumwsqr
        result.sumwsqrsq = self.sumwsqrsq + other.sumwsqrsq

        return result


class Estimator(base.Estimator):
    _impl: EstimatorImpl

    def __init__(self, wmin: float = 0, wmax: float = inf):
        self._impl = EstimatorImpl(wmin, wmax)

    def add_example(self, p_log: float, r: float, p_pred: float, count: float = 1.0) -> None:
        self._impl.add(p_pred / p_log, r, count)

    def get(self) -> Optional[float]:
        return self._impl.get()

    def __add__(self, other: 'Estimator') -> 'Estimator':
        result = Estimator()
        result._impl = self._impl + other._impl
        return result


class Interval(base.Interval):
    _impl: IntervalImpl

    def __init__(self, wmin: float = 0, wmax: float = inf, rmin: float = 0, rmax: float = 1, empirical_r_bounds = False):
        self._impl = IntervalImpl(wmin, wmax, rmin, rmax, empirical_r_bounds)

    def add_example(self, p_log: float, r: float, p_pred: float, count: float = 1.0) -> None:
        self._impl.add(p_pred / p_log, r, count)

    def get(self, alpha: float = 0.05, atol: float = 1e-9) -> List[Optional[float]]:
        return self._impl.get(alpha, atol)

    def __add__(self, other: 'Interval') -> 'Interval':
        result = Interval()
        result._impl = self._impl + other._impl
        return result



""" Interface for implementation of contextual bandit estimators """

from abc import ABC, abstractmethod
from typing import List, Optional


class Estimator(ABC):
	""" Interface for implementation of contextual bandit estimators """

	@abstractmethod
	def add_example(self, p_log: float, r: float, p_pred: float, count: float) -> None:
		""" 
		Args:
			p_log: probability of the logging policy
			r: reward for choosing an action in the given context
			p_pred: predicted probability of making decision
			count: weight
		"""
		...

	@abstractmethod
	def get(self) -> Optional[float]:
		""" Calculates the selected estimator
		
		Returns:
			The estimator value
		"""
		...


class Interval(ABC):
	""" Interface for implementation of contextual bandit estimators interval """

	@abstractmethod
	def add_example(self, p_log: float, r: float, p_pred: float, count: float) -> None:
		""" 
		Args:
			p_log: probability of the logging policy
			r: reward for choosing an action in the given context
			p_pred: predicted probability of making decision
			count: weight
		"""
		...

	@abstractmethod
	def get(self, alpha: float) -> List[Optional[float]]:
		""" Calculates the CI
		Args:
			alpha: alpha value
		Returns:
			Returns the confidence interval as list[float]
		"""
		...



import sys

def update_progress(current, total=None, prefix=''):
    if total:
        barLength = 50 # Length of the progress bar
        progress = current/total
        block = int(barLength*progress)
        text = "\r{}Progress: [{}] {:.1f}%".format(prefix, "#"*block + "-"*(barLength-block), progress*100)
    else:
        text = "\r{}Iter: {}".format(prefix, current)
    sys.stdout.write(text)
    sys.stdout.flush()
    return len(text)-1      # return length of text string not counting the initial \r

def json_cooked(x):
    #################################
    # Optimized version based on expected structure:
    # {"_label_cost":0,"_label_probability":0.01818182,"_label_Action":9,"_labelIndex":8,"Timestamp":"2017-10-24T00:00:15.5160000Z","Version":"1","EventId":"fa68cd9a71764118a635fd3d7a908634","a":[9,11,3,1,6,4,10,5,7,8,2],"c":{"_synthetic":false,"User":{"_age":0},"Geo":{"country":"United States","_countrycf":"8","state":"New York","city":"Springfield Gardens","_citycf":"8","dma":"501"},"MRefer":{"referer":"http://www.complex.com/"},"OUserAgent":{"_ua":"Mozilla/5.0 (iPad; CPU OS 10_3_2 like Mac OS X) AppleWebKit/603.2.4 (KHTML, like Gecko) Version/10.0 Mobile/14F89 Safari/602.1","_DeviceBrand":"Apple","_DeviceFamily":"iPad","_DeviceIsSpider":false,"_DeviceModel":"iPad","_OSFamily":"iOS","_OSMajor":"10","_OSPatch":"2","DeviceType":"Tablet"},"_multi":[{"
    # {"_label_cost":0,"_label_probability":1,"_label_Action":1,"_labelIndex":0,"_deferred":true,"Timestamp":"2018-10-25T00:01:31.1780000Z","Version":"1","EventId":"28EF7EE0B9CF4E319696CB812973F0B3","DeferredAction":true,"a":[1],"c":{"Global":{"SLOT":"1"},"Request":{"DISPLOC":"BR"},"Profile":{"_REQASID":"F2A3670E24F94646983BEB3EB35CB0C9","COUNTRY":"BR","FPIGM":"FALSE","FPNIGM":"FALSE","FB":"FALSE"},"_multi":[{"Action":{"constant":1,"PayloadID":"425039838"}}]},"p":[1.000000],"VWState":{"m":"DF30E6A3648947E69EE6B0816BF42640/1A5CD300DFE546B7B29A11EB70980809"}}
    # {"_label_cost":0,"_label_probability":0.833333015,"_label_Action":6,"_labelIndex":5,"o":[{"EventId":"A3E5ADF82D3A4BD5A5161FFC19C95DBB","DeferredAction":false}],"Timestamp":"2018-10-25T00:00:00.3960000Z","Version":"1","EventId":"A3E5ADF82D3A4BD5A5161FFC19C95DBB","DeferredAction":true,"a":[6,2,4,5,1,3],"c":{"Global":{"SLOT":"2"},"Request":{"DISPLOC":"US"},"Profile":{"_REQASID":"70A03B1FD0CA4B5A89669637DD448161","COUNTRY":"US","FPIGM":"FALSE","FPNIGM":"FALSE","FB":"FALSE","F1":0.0,"F2":0.0,"F3":0.00157480315,"F4":0.00182481752,"F5":0.0,"F6":0.00136892539,"F7":0.0,"F8":-1.0,"F9":0.0434782609,"F11":0.0,"F12":0.0,"F13":0.0,"F14":1.0,"F15":1.0,"F16":0.0,"F17":2.0,"F18":0.0,"F19":0.0,"F20":1.0
    # Assumption: "Version" value is 1 digit string
    #
    # Performance: 4x faster than Python JSON parser js = json.loads(x.strip())
    #################################
    ind1 = x.find(b',',16)              # equal to: x.find(',"_label_prob',16)
    ind2 = x.find(b',',ind1+23)         # equal to: x.find(',"_label_Action',ind1+23)
    ind3 = x.find(b',"T',ind2+34)       # equal to: x.find(',"Timestamp',ind2+34)
    ind7 = x.find(b',"a"',ind3+60)
    ind8 = x.find(b']',ind7+7)          # equal to: x.find('],"c',ind7+8)

    data = {}
    data['cost'] = x[15:ind1]                   # len('{"_label_cost":') = 15
    data['p'] = float(x[ind1+22:ind2])          # len(',"_label_probability":') = 22
    data['a_vec'] = x[ind7+6:ind8].split(b',')  # len(',"a":[') = 6
    data['a'] = int(data['a_vec'][0])
    data['num_a'] = len(data['a_vec'])
    data['skipLearn'] = b'"_skipLearn":true' in x[ind2+34:ind3] # len('"_label_Action":1,"_labelIndex":0,') = 34
            
    return data

def json_cooked_continuous_actions(x):
    #################################
    # Optimized version based on expected structure:
    # {"_label_ca":{"cost":0,"pdf_value":0.0181818,"action":185.5},"Timestamp":"2017-10-24T00:00:15.5160000Z","Version":"1","EventId":"fa68cd9a71764118a635fd3d7a908634","c":{}}"
    # Assumption: "Version" value is 1 digit string
    #
    # Performance: 4x faster than Python JSON parser js = json.loads(x.strip())
    #################################
    ind1 = x.find(b',',22)              # equal to: x.find(',"pdf_value',16)
    ind2 = x.find(b',',ind1+13)         # equal to: x.find(',"action',ind1+23)
    ind3 = x.find(b'}',ind2+10)
    ind4 = x.find(b',"T',ind3+34)       # equal to: x.find(',"Timestamp',ind2+34)

    data = {}
    data['cost'] = float(x[21:ind1])                   # len('{"_label_ca":"cost":') = 21
    data['p'] = float(x[ind1+13:ind2])          # len(',"pdf_value":') = 13
    data['a'] = float(x[ind2+10:ind3])
    data['skipLearn'] = b'"_skipLearn":true' in x

    return data